# -----------------------------------------------------------------------
#
# ff - a tool to search the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

from . import NoData
from .type import *

__all__ = ["NoData", "Type", "String", "Number", "FileType", "Mode",
        "Size", "Time", "Duration", "Boolean", "ListOfStrings", "Plugin"]


class Plugin:
    """Base class for all plugins.
    """

    use_cache = False
    attributes = {}

    def __init_subclass__(cls):
        super().__init_subclass__()

        cls.name = cls.__name__.lower()

    @classmethod
    def initialize(cls, module_source, module_path, module_tag):
        """Initialize the Plugin class with basic information.
        """
        cls.module_source = module_source
        cls.module_path = module_path
        cls.module_tag = module_tag
        cls.sql_table_name = f"plugin_{cls.name}_{cls.module_tag}"

    def __init__(self):
        self.setup()

    def setup(self):
        """Setup the Plugin class. This is called once for every plugin, but
           only if it will be actually used. This is mainly supposed to be used
           for importing third-party modules that may or may not be installed
           on the system.
        """

    def can_handle(self, entry):
        """Return True if this plugin can handle this particular Entry object.
           This is supposed to reduce calls to process() and minimize the
           number of needlessly cached entries.
        """
        raise NotImplementedError

    def process(self, entry):
        """Generate key-value pairs or return a dict with the values that were
           "extracted" from the entry. The set of keys must correspond with the
           ones from the attributes dict, but it is not necessary that all the
           keys actually have values. Raise a NoData exception or return an
           empty sequence if the plugin cannot extract the required attributes
           properly.
        """
        raise NotImplementedError



# -----------------------------------------------------------------------
#
# ff - a tool to search the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import sys
import pstats
import cProfile

from libff import EX_OK, EX_USAGE, EX_PROCESS, EX_SUBPROCESS, Directory
from libff.walk import FilesystemWalker
from libff.cache import Cache, NullCache
from libff.entry import StartDirectory
from libff.field import ExecFields, SortFields, CountFields, OutputFields, \
    ExecBatchFields
from libff.filter import Matcher, Excluder
from libff.ignore import GitIgnore
from libff.parser import ParserError
from libff.context import Context
from libff.registry import Registry
from libff.arguments import parse_arguments
from libff.processing import ImmediateExecProcessing, \
    CollectiveExecProcessing, ImmediateConsoleProcessing, \
    CollectiveConsoleProcessing

if __name__ == "__main__":
    # Set up the Context object that is passed to all components.
    context = Context()

    # Parse command-line arguments.
    # Once we're on Python >=3.8 we can use context.args = (args := parse_arguments(context)).
    args = parse_arguments(context)
    context.args = args
    context.setup()

    # Setup the sqlite3 metadata cache.
    if args.cache is None:
        CacheClass = NullCache
    else:
        CacheClass = Cache
    cache = CacheClass(context)
    context.cache = cache

    # Register all types and metadata plugins.
    registry = Registry(context)
    context.registry = registry

    # Replace the special value "file" in -o/--output by a list of all file
    # attributes.
    if "file" in args.output:
        pos = args.output.index("file")
        args.output.pop(pos)
        args.output = args.output[:pos] + list(registry.get_file_attributes()) + args.output[pos:]

    # Turn all options that take fields as arguments in to Fields objects.
    args.output = OutputFields(context, args.output)
    if args.sort is not None:
        args.sort = SortFields(context, args.sort)
    if args.count is not None:
        args.count = CountFields(context, args.count)
    if args.exec is not None:
        args.exec = ExecFields(context, args.exec)
    if args.exec_batch is not None:
        args.exec_batch = ExecBatchFields(context, args.exec_batch)

    if args.list_attributes:
        registry.print_attributes(with_operators=True)
        raise SystemExit(EX_OK)

    elif args.list_plugins:
        registry.print_plugins()
        raise SystemExit(EX_OK)

    elif args.list_types:
        registry.print_types()
        raise SystemExit(EX_OK)

    elif args.help is not None:
        registry.print_help(args.help)
        raise SystemExit(EX_OK)

    try:
        # Setup the matcher from the command line arguments.
        matcher = Matcher(context, args.expressions)
        context.matcher = matcher
    except ParserError as exc:
        context.error(f"unable to parse expressions: {exc}", EX_USAGE)

    # Setup the excluder from the command line arguments.
    excluder = Excluder(context, args.exclude)
    context.excluder = excluder

    # Set up processing. There are two modes of operation: Either collect all
    # results and process them in their entirety, or process results
    # immediately as they come in.
    if args.sort or args.count or args.limit or args.exec_batch or args.json == "json":
        if args.exec or args.exec_batch:
            processing = CollectiveExecProcessing(context)
            context.processing = processing
        else:
            processing = CollectiveConsoleProcessing(context)
            context.processing = processing
    elif args.exec:
        processing = ImmediateExecProcessing(context)
        context.processing = processing
    else:
        processing = ImmediateConsoleProcessing(context)
        context.processing = processing

    # Set up the filesystem walker.
    walker = FilesystemWalker(context)
    context.walker = walker

    if __debug__:
        context.debug("info", f"Using {context.processing.__class__.__name__}.")

        # Show some debug output.
        context.debug("info", "Directories to search:")
        for directory in args.directories:
            context.debug("info",
                    f"    {directory if directory != '.' else os.path.abspath(directory)}")

        if not excluder.is_empty():
            context.debug("info", "Exclude Sequence:")
            for line in excluder.parser.format():
                context.debug("info", "  " + line)

        if not matcher.is_empty():
            context.debug("info", "Test Sequence:")
            for line in matcher.parser.format():
                context.debug("info", "  " + line)

    # Preload the in-queue with the path arguments.
    for path in args.directories:
        if args.absolute_path:
            path = os.path.abspath(path)

        if __debug__ and args.ignore_parent_ignorefiles:
            # Don't use ignore files from parent directories.
            ignores = []
        else:
            # Find ignore files in parent directories.
            if args.ignored:
                ignores = GitIgnore.from_directory(context, path)
            else:
                ignores = []

        walker.put([Directory(StartDirectory(args, path), "", ignores)])

    # Start processing.
    try:
        if __debug__ and args.profile:
            # Start profiling. No multiprocessing is involved, Walker.loop() is
            # run in the main thread.
            profiler = cProfile.Profile()
            profiler.enable()
            walker.loop(0)
        else:
            walker.start_processes()

        # Collect Entry objects from the Walker.
        processing.loop()

        # Finalize and put out the result if --sort or --exec-batch is
        # involved.
        processing.finalize()

        if __debug__ and args.profile:
            # Stop profiling and print the statistics.
            profiler.disable()
            stats = pstats.Stats(profiler, stream=sys.stderr)
            stats.sort_stats("cumulative")
            stats.print_stats(.1)

    except KeyboardInterrupt:
        # Stop all processes immediately.
        context.stop()
        if __debug__:
            raise
        else:
            raise SystemExit("keyboard interrupt")

    except BrokenPipeError:
        context.stop()

    context.close()
    walker.close()
    processing.close()

    if __debug__:
        hits = context.cache_hits.value
        misses = context.cache_misses.value
        if hits or misses:
            context.debug("cache", f"Cache stats: {hits} hits, {misses} misses")
        else:
            context.debug("cache", "Cache was not used")

    if context.exitcode == EX_SUBPROCESS:
        context.error("One or more --exec/--exec-batch commands had errors", EX_SUBPROCESS)
    elif context.exitcode == EX_PROCESS:
        context.error("One or more ff processes had unrecoverable errors! "\
                "Result is probably incomplete!", EX_PROCESS)



# -----------------------------------------------------------------------
#
# ff - a tool to search the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import textwrap
import itertools

from . import OUTPUT_WIDTH


class Table(list):
    """A simple class for tabular output.
    """

    def __init__(self, column_names, padding="   ", indent="    ", wrap_last_column=False):
        super().__init__()
        self.column_names = column_names
        self.padding = padding
        self.indent = indent
        self.wrap_last_column = wrap_last_column

        self.column_widths = []
        for name in column_names:
            self.column_widths.append(len(name))

    def calculate_last_column_length(self):
        """Calculate the length of the last column if it supposed to be
           wrapped.
        """
        width = OUTPUT_WIDTH
        width -= len(self.indent)
        width -= self.calculate_padding()
        width -= sum(self.column_widths[:-1])
        return max(25, width)

    def calculate_padding(self):
        """Calculate the total length of the padding between the columns.
        """
        return len(self.padding) * (len(self.column_widths) - 1)

    def add(self, row):
        """Add a new row of values.
        """
        if self.wrap_last_column:
            row = row[:-1] + [" ".join(row[-1].strip().split())]

        for i, value in enumerate(row):
            self.column_widths[i] = max(len(value), self.column_widths[i])

        self.append(row)

    def format(self, row):
        """Format a row and generate a list of lines for output. Optionally,
           wrap the last column.
        """
        if self.wrap_last_column:
            for new_row in itertools.zip_longest(
                    *([value] for value in row[:-1]),
                    textwrap.wrap(row[-1], width=self.calculate_last_column_length()),
                    fillvalue=""):
                yield new_row
        else:
            yield row

    def print_row(self, row):
        """Format a row and print it to standard output.
        """
        for new_row in self.format(row):
            print(self.indent, end="")
            for value, width in zip(new_row[:-1], self.column_widths[:-1]):
                print(value.ljust(width), end=self.padding)
            print(new_row[-1])

    def print_separator(self):
        """Print the separator between header and values.
        """
        width = min(OUTPUT_WIDTH - len(self.indent),
                sum(self.column_widths) + self.calculate_padding())
        print(self.indent + "-" * width)

    def print(self):
        """Print the table to standard output.
        """
        self.print_row(self.column_names)
        self.print_separator()
        for row in self:
            self.print_row(row)



# -----------------------------------------------------------------------
#
# ff - a tool to search the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import queue
import shutil
import threading
import collections
import multiprocessing

from .__version__ import __version__

__copyright__= f"""ff {__version__}
Copyright (C) 2020 Lars Gustäbel <https://github.com/gustaebel/ff>
License GPLv3+: GNU GPL version 3 or later <https://gnu.org/licenses/gpl.html>.

This is free software: you are free to change and redistribute it.
There is NO WARRANTY, to the extent permitted by law."""

NOTSET = object()

TIMEOUT = 0.01
MAX_CPU = multiprocessing.cpu_count()


class NoData(Exception):
    """Exception for when a plugin cannot process an entry.
    """

EX_OK = 0
EX_USAGE = 1
EX_SUBPROCESS = 2
EX_PROCESS = 3
EX_BAD_PLUGIN = 10
EX_BAD_ATTRIBUTE = 11
EX_EXPRESSION = 12

OUTPUT_WIDTH = shutil.get_terminal_size((100, 100))[0] - 2


Directory = collections.namedtuple("Directory", "start relpath ignores")
Entries = collections.namedtuple("Entries", "parent entries ignores")

class Attribute(collections.namedtuple("Attribute", "plugin name")):
    """An Attribute namedtuple that contains the plugin name and the attribute
       name.
    """

    def __str__(self):
        return f"{self.plugin}.{self.name}"


# pylint:disable=too-few-public-methods
class BaseClass:
    """This class can be used as the base for each class that depends on
       Context and access to args, registry, etc. It offers shortcuts to the
       each of these components as self.component instead of
       self.context.component without using properties. Not exactly a matter of
       life and death but nice nevertheless.
    """

    component_names = set(["args", "registry", "excluder", "matcher", "walker"])

    def __init__(self, context):
        self.context = context

    def __getattr__(self, name):
        try:
            attr = getattr(self.context, name)
        except AttributeError:
            raise AttributeError(f"{self.__class__.__name__} object has no attribute {name!r}")

        if name in self.component_names:
            setattr(self, name, attr)
        return attr



# -----------------------------------------------------------------------
#
# ff - a tool to search the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

from libff.plugin import *


class Grep(Plugin):
    """Plugin that gives access to the lines of a text file.
    """

    use_cache = False
    attributes = [
        ("lines", ListOfStrings, "The lines of the file.")
    ]

    def can_handle(self, entry):
        return entry.text

    def process(self, entry):
        with open(entry.path, errors="surrogateescape") as lines:
            yield "lines", list(lines)



# -----------------------------------------------------------------------
#
# ff - a tool to search the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import tarfile

from libff.plugin import *


class Tar(Plugin):
    """Extract information from tar archives.
    """

    use_cache = True
    attributes = [
        ("members", ListOfStrings, "The list of file names that are stored in a .tar archive.")
    ]

    extensions = (
            ".tar",
            ".tar.gz", ".taz", ".tgz", ".tpz",
            ".tar.bz2", ".tb2", ".tbz", ".tbz2", ".tz2",
            ".tar.lz", ".tar.lzma", ".tlz",
            ".tar.xz", ".txz")

    def can_handle(self, entry):
        return entry.name.endswith(self.extensions)

    def process(self, entry):
        try:
            with tarfile.open(entry.path) as tar:
                yield "members", set(name for name in tar.getnames())
        except (OSError, EOFError):
            raise NoData



# -----------------------------------------------------------------------
#
# ff - a tool to search the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import subprocess

from libff.plugin import *


class Git(Plugin):
    """Extract information from files that are inside a git(1) repository.
    """

    use_cache = True
    attributes = [
        ("tracked", Boolean, "Whether the file is tracked by a git repository.")
    ]

    def check_tracked(self, entry):
        """Return True if the Entry object is under version control.
        """
        return subprocess.run(["git", "ls-files", "--error-unmatch", entry.basename],
                cwd=entry.dirname, check=False,
                stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL).returncode == 0

    def can_handle(self, entry):
        return entry.is_dir() or entry.is_file() or entry.is_symlink()

    def process(self, entry):
        try:
            yield "tracked", self.check_tracked(entry)
        except OSError:
            raise NoData



# Put this file in ~/.ff and rename it.

from libff.plugin import *

# IMPORTANT: Do not import third-party modules here, that may not
# be installed on the system, do this in Plugin.setup().


class Name(Plugin):
    # Give the Plugin a descriptive name that does not yet exist.

    # If the code in the process() method takes a lot of time to produce the
    # results, and under the condition that the result would not change between
    # different calls to process(), you can turn on caching here. A once cached
    # result is discarded as soon as the associated entry's modification time
    # changes or the code of this module is altered.
    use_cache = False

    # Here you tell ff(1) which attributes your plugin will provide.
    # `attributes` is a list of 3-element-tuples: (name, type, description).
    # `type` must be a subclass of Type. A Type's job is for example to
    # describe which operators the value supports, and to format the value
    # for output or to a human readable form. You can create your own Types if
    # needed, see libff/type.py for details.
    #
    # Available Types are:
    #
    #    String             a generic string
    #    ListOfStrings      a list of strings
    #    Number             a number greater or equal to 0
    #    Size               a size in bytes
    #    Time               a point in time in seconds
    #    Duration           a duration in seconds
    #    Boolean            a true or false value
    #
    attributes = [
        ("name", String, "The name of the file."),
        ("size", Size, "The size of the file.")
    ]

    @classmethod
    def setup(cls):
        # Optional.
        #
        # This classmethod is called once before the Plugin is about to be
        # used. It is supposed to be used for setup code, mainly for importing
        # modules that may or may not be installed.
        #
        #   @classmethod
        #   def setup(cls):
        #       global module
        #       import module

    def can_handle(self, entry):
        # Required.
        #
        # Return True if this plugin can handle this type of file. There are
        # many ways possible for you to do this: based on an attribute of the
        # Entry object (e.g. file type, extension, mime type etc.) or by
        # reading a magic number from the file.
        # This check is supposed to reduce the number the process() method is
        # called. If use_cache is active, each result from process() will be
        # cached, even if it is empty, so a properly implemented can_handle()
        # method will also reduce the load on the cache and the number of
        # needlessly cached entries.
        return entry.is_file()

    def process(self, entry):
        # Required.
        #
        # Extract information from the entry to produce values for the
        # attributes described above, e.g. by examining the contents of the
        # file. `entry` is a an Entry object from libff/entry.py, see there
        # for attributes and properties.
        # On success, either return a dictionary with the values or yield
        # key-value pairs. On failure, i.e. if the file information cannot be
        # extracted, raise a NoData exception or return an empty dictionary.
        yield "name", entry.name
        yield "size", entry.size



# -----------------------------------------------------------------------
#
# ff - a tool to search the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import sys
import queue
import textwrap
import threading
import traceback
import multiprocessing
import multiprocessing.sharedctypes

from . import TIMEOUT, OUTPUT_WIDTH


class Context:
    """The Context class contains a global state and is globally accessible so that
       all components find all other components.
    """
    # pylint:disable=too-many-instance-attributes

    warnings_emitted = set()

    def __init__(self):
        # These are added later.
        self.args = None
        self.cache = None
        self.registry = None
        self.matcher = None
        self.excluder = None
        self.walker = None
        self.processing = None

        self.stop_queue = None
        self.barrier = None
        self.exitcode_object = None

        self.cache_hits = None
        self.cache_misses = None

    def setup(self):
        """Set up multiprocessing and inter-process communication.
        """
        # We use a Queue object to emulate a stop event, because
        # multiprocessing.Event did not work the way I expected it.
        self.stop_queue = multiprocessing.Queue(self.args.jobs)

        # Create a Barrier with the number of processes plus one for the main
        # process.
        if __debug__ and self.args.profile:
            num_jobs = 1
        else:
            num_jobs = self.args.jobs + 1
        self.barrier = multiprocessing.Barrier(num_jobs)

        # The exitcode object provides a way for --exec and --exec-batch
        # processes to feed back errors to the main process.
        self.exitcode_object = multiprocessing.sharedctypes.RawValue("i", 0)

        # Provide a shared counter for cache hits and misses.
        self.cache_hits = multiprocessing.Value("L")
        self.cache_misses = multiprocessing.Value("L")

    def close(self):
        """Close the Context and clean up.
        """
        self.stop_queue.close()

    def set_exitcode(self, exitcode):
        """Set the shared resource 'exitcode' to the value that will
           be used as the main exit code.
        """
        self.exitcode_object.value = exitcode

    @property
    def exitcode(self):
        """The exitcode to use as the main exit code.
        """
        return self.exitcode_object.value

    def barrier_wait(self, timeout=TIMEOUT):
        """Wait for all processes to reach the Barrier. If all processes do
           that means that there is no more input and processing is finished.
        """
        # Use a Barrier object to determine if all processes are currently
        # waiting for input, which means the operation is finished.
        try:
            self.barrier.wait(timeout)
        except threading.BrokenBarrierError:
            if self.is_stopping():
                # Stop immediately without waiting for all processes to finish.
                return True
            else:
                self.barrier.reset()
                return False
        else:
            return True

    def idle_processes(self):
        """Return the number of idle processes, i.e. processes that are waiting
           for the barrier because there is nothing in the queue.
        """
        return self.barrier.n_waiting

    def is_stopping(self):
        """Return True all Processes are supposed to stop.
        """
        try:
            self.stop_queue.get_nowait()
        except queue.Empty:
            return False
        else:
            return True

    def stop(self):
        """Stop all processes immediately.
        """
        for _ in range(self.args.jobs):
            self.stop_queue.put(None)

    def wrap(self, category, message):
        """Format and wrap the text of a message for output.
        """
        return textwrap.fill(category + ": " + str(message), width=OUTPUT_WIDTH,
                subsequent_indent=" " * (len(category) + 2))

    def warning(self, message, tag=None):
        """Print a warning to standard error output. If tag is not None hide
           the warning if there has already been another warning with that tag.
        """
        if tag is not None:
            if tag in self.warnings_emitted:
                return
            else:
                self.warnings_emitted.add(tag)

        print(self.wrap("WARNING", message), file=sys.stderr)

    def error(self, message, exitcode):
        """Print an error message to standard error output and exit.
        """
        print(self.wrap("ERROR", message), file=sys.stderr)
        if exitcode is not None:
            raise SystemExit(exitcode)

    def info(self, message):
        """Print an info message to standard error output.
        """
        print(self.wrap("INFO", message), file=sys.stderr)

    if __debug__:
        def debug(self, category, message):
            """Print a debug message to standard error output if --debug is
               switched on.
            """
            if self.args.debug is None or category in self.args.debug:
                print(f"DEBUG:{category}: {message}", file=sys.stderr)

        def debug_proc(self, index, message):
            """Print a debug message to standard error output if --debug-processes
               is switched on.
            """
            self.debug("mp", f"#{index:02d} {message}")

    def exception(self, message, exitcode=None):
        """Print an error message and a traceback to standard error output and
           exit.
        """
        print(self.wrap("INTERNAL", message + ":"), file=sys.stderr)
        traceback.print_exc(file=sys.stderr)
        if exitcode is not None:
            raise SystemExit(exitcode)
